/*
 * jz4740 LCD controller configuration.
 *
 * Copyright (C) Xiangfu Liu <xiangfu@sharism.cc>
 * Copyright (C) 2015, 2016, 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "sdram.h"
#include "jzlcd.h"
#include "cpu.h"
#include "board.h"

extern vidinfo_t panel_info;

/* Useful alignment operations. */

static inline void align2(uint32_t *n)
{
	*n = (((*n)+1)>>1)<<1;
}

static inline void align4(uint32_t *n)
{
	*n = (((*n)+3)>>2)<<2;
}

static inline void align8(uint32_t *n)
{
	*n = (((*n)+7)>>3)<<3;
}



/* Register operations. */

static inline uint32_t lcd_ctrl_get(vidinfo_t *vid, uint32_t reg)
{
	return REG32(vid->lcd + reg);
}

static inline void lcd_ctrl_set(vidinfo_t *vid, uint32_t reg, uint32_t value)
{
	REG32(vid->lcd + reg) = value;
}



/* Configuration operations. */

/* Return the number of panels available. */

static uint8_t lcd_get_panels(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	return ((jzfb->cfg & MODE_MASK) == MODE_STN_MONO_DUAL) ||
		((jzfb->cfg & MODE_MASK) == MODE_STN_COLOR_DUAL) ? 2 : 1;
}

/* Calculate and return the pixel clock frequency. */

static uint32_t lcd_get_pixel_clock(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	uint32_t pclk, width_cycles, mode = jzfb->cfg & MODE_MASK;

	/*
	Serial mode: 3 pixel clock cycles per pixel (one per channel).
	Parallel mode: 1 pixel clock cycle per pixel.
	*/

	if (mode == MODE_8BIT_SERIAL_TFT)
		width_cycles = jzfb->w * 3;
	else
		width_cycles = jzfb->w;

	/* Derive pixel clock from frame clock. */

	pclk = jzfb->fclk *
		(width_cycles + jzfb->hsw + jzfb->elw + jzfb->blw) *
		(jzfb->h + jzfb->vsw + jzfb->efw + jzfb->bfw);

	if ((mode == MODE_STN_COLOR_SINGLE) || (mode == MODE_STN_COLOR_DUAL))
		pclk = (pclk * 3);

	if ((mode == MODE_STN_COLOR_SINGLE) || (mode == MODE_STN_COLOR_DUAL) ||
	    (mode == MODE_STN_MONO_SINGLE) || (mode == MODE_STN_MONO_DUAL))
		pclk = pclk >> ((jzfb->cfg & STN_DAT_PINMASK) >> 4);

	if ((mode == MODE_STN_COLOR_DUAL) || (mode == MODE_STN_MONO_DUAL))
		pclk >>= 1;

	return pclk;
}



/* Functions returning region sizes. */

static uint32_t lcd_get_size(vidinfo_t *vid)
{
	/* Lines must be aligned to a word boundary. */
	uint32_t line_length = ALIGN((vid->jz_fb->w * vid->jz_fb->bpp) / 8, sizeof(uint32_t));
        return line_length * vid->jz_fb->h;
}

static uint32_t lcd_get_aligned_size(vidinfo_t *vid)
{
	/* LCD_CTRL_BST_16 requires 16-word alignment. */
        return ALIGN(lcd_get_size(vid), 16 * sizeof(uint32_t));
}

static uint32_t lcd_get_min_size(vidinfo_t *vid)
{
	/* Lines must be aligned to a word boundary. */
	uint32_t line_length = ALIGN((vid->jz_fb->w * 32) / 8, sizeof(uint32_t));
        return line_length * vid->jz_fb->h;
}

static uint32_t lcd_get_aligned_min_size(vidinfo_t *vid)
{
	/* LCD_CTRL_BST_16 requires 16-word alignment. */
        return ALIGN(lcd_get_min_size(vid), 16 * sizeof(uint32_t));
}

static uint32_t lcd_get_palette_size(vidinfo_t *vid)
{
	/* Get a collection of two-byte entries, one per colour. */

	if (vid->jz_fb->bpp < 12)
		return (1 << (vid->jz_fb->bpp)) * sizeof(uint16_t);
	else
		return 0;
}

static uint32_t lcd_get_aligned_palette_size(vidinfo_t *vid)
{
	/* LCD_CTRL_BST_16 requires 16-word alignment. */
	return ALIGN(lcd_get_palette_size(vid), 16 * sizeof(uint32_t));
}

static uint32_t lcd_get_descriptors_size()
{
	return 3 * sizeof(struct jz_fb_dma_descriptor);
}

static uint32_t lcd_get_total_size(vidinfo_t *vid)
{
	uint32_t size = lcd_get_aligned_size(vid) * lcd_get_panels(vid);
	uint32_t min_size = lcd_get_aligned_min_size(vid);

        /* Round up to nearest full page, or MMU section if defined. */
	return ALIGN((size >= min_size ? size : min_size) + lcd_get_aligned_palette_size(vid) + lcd_get_descriptors_size(), PAGE_SIZE);
}



/* Functions returning addresses of each data region. */

static uint32_t lcd_get_palette(vidinfo_t *vid)
{
        /* Use memory at the end of the allocated region for the palette. */

	return get_memory_size() - lcd_get_aligned_palette_size(vid);
}

static uint32_t lcd_get_descriptors(vidinfo_t *vid)
{
        /* Use memory before the palette for the descriptor array. */

	return lcd_get_palette(vid) - lcd_get_descriptors_size();
}

static uint32_t lcd_get_framebuffer(uint16_t panel, vidinfo_t *vid)
{
        /* Framebuffers for panels are allocated at the start of the region. */

        return get_memory_size() - lcd_get_total_size(vid) + (panel * lcd_get_aligned_size(vid));
}



/* Initialisation functions. */

static uint32_t jz_lcd_stn_init(uint32_t stnH, vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	uint32_t val = 0;

	switch (jzfb->bpp) {
		case 1:
		/* val |= LCD_CTRL_PEDN; */
		case 2:
		val |= LCD_CTRL_FRC_2;
		break;

		case 4:
		val |= LCD_CTRL_FRC_4;
		break;

		case 8:
		default:
		val |= LCD_CTRL_FRC_16;
		break;
	}

	switch (jzfb->cfg & STN_DAT_PINMASK) {
		case STN_DAT_PIN1:
		/* Do not adjust the hori-param value. */
		break;

		case STN_DAT_PIN2:
		align2(&jzfb->hsw);
		align2(&jzfb->elw);
		align2(&jzfb->blw);
		break;

		case STN_DAT_PIN4:
		align4(&jzfb->hsw);
		align4(&jzfb->elw);
		align4(&jzfb->blw);
		break;

		case STN_DAT_PIN8:
		align8(&jzfb->hsw);
		align8(&jzfb->elw);
		align8(&jzfb->blw);
		break;
	}

	lcd_ctrl_set(vid, LCD_VSYNC, jzfb->vsw);
	lcd_ctrl_set(vid, LCD_HSYNC, ((jzfb->blw + jzfb->w) << 16) | (jzfb->blw + jzfb->w + jzfb->hsw));

	/* Screen setting */
	lcd_ctrl_set(vid, LCD_VAT, ((jzfb->blw + jzfb->w + jzfb->hsw + jzfb->elw) << 16) | (stnH + jzfb->vsw + jzfb->bfw + jzfb->efw));
	lcd_ctrl_set(vid, LCD_DAH, (jzfb->blw << 16) | (jzfb->blw + jzfb->w));
	lcd_ctrl_set(vid, LCD_DAV, stnH);

	/* AC BIAs signal */
	lcd_ctrl_set(vid, LCD_PS, stnH+jzfb->vsw+jzfb->efw+jzfb->bfw);

	return val;
}

static void jz_lcd_tft_init(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;

	lcd_ctrl_set(vid, LCD_VSYNC, jzfb->vsw);
	lcd_ctrl_set(vid, LCD_HSYNC, jzfb->hsw);
	lcd_ctrl_set(vid, LCD_DAV, ((jzfb->vsw + jzfb->bfw) << 16) | (jzfb->vsw + jzfb->bfw + jzfb->h));
	lcd_ctrl_set(vid, LCD_DAH, ((jzfb->hsw + jzfb->blw) << 16) | (jzfb->hsw + jzfb->blw + jzfb->w));
	lcd_ctrl_set(vid, LCD_VAT, (((jzfb->blw + jzfb->w + jzfb->elw + jzfb->hsw)) << 16) |
		(jzfb->vsw + jzfb->bfw + jzfb->h + jzfb->efw));
}

static void jz_lcd_samsung_init(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	uint32_t pclk = lcd_get_pixel_clock(vid);
	uint32_t total, tp_s, tp_e, ckv_s, ckv_e;
	uint32_t rev_s, rev_e, inv_s, inv_e;

	jz_lcd_tft_init(vid);

	total = jzfb->blw + jzfb->w + jzfb->elw + jzfb->hsw;
	tp_s = jzfb->blw + jzfb->w + 1;
	tp_e = tp_s + 1;
	/* ckv_s = tp_s - jz_clocks.pixclk/(1000000000/4100); */
	ckv_s = tp_s - pclk/(1000000000/4100);
	ckv_e = tp_s + total;
	rev_s = tp_s - 11;	/* -11.5 clk */
	rev_e = rev_s + total;
	inv_s = tp_s;
	inv_e = inv_s + total;
	lcd_ctrl_set(vid, LCD_CLS, (tp_s << 16) | tp_e);
	lcd_ctrl_set(vid, LCD_PS, (ckv_s << 16) | ckv_e);
	lcd_ctrl_set(vid, LCD_SPL, (rev_s << 16) | rev_e);
	lcd_ctrl_set(vid, LCD_REV, (inv_s << 16) | inv_e);
	jzfb->cfg |= STFT_REVHI | STFT_SPLHI;
}

static void jz_lcd_sharp_init(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	uint32_t total, cls_s, cls_e, ps_s, ps_e;
	uint32_t spl_s, spl_e, rev_s, rev_e;

	jz_lcd_tft_init(vid);

	total = jzfb->blw + jzfb->w + jzfb->elw + jzfb->hsw;
	spl_s = 1;
	spl_e = spl_s + 1;
	cls_s = 0;
	cls_e = total - 60;	/* > 4us (pclk = 80ns) */
	ps_s = cls_s;
	ps_e = cls_e;
	rev_s = total - 40;	/* > 3us (pclk = 80ns) */
	rev_e = rev_s + total;
	jzfb->cfg |= STFT_PSHI; 
	lcd_ctrl_set(vid, LCD_SPL, (spl_s << 16) | spl_e);
	lcd_ctrl_set(vid, LCD_CLS, (cls_s << 16) | cls_e);
	lcd_ctrl_set(vid, LCD_PS, (ps_s << 16) | ps_e);
	lcd_ctrl_set(vid, LCD_REV, (rev_s << 16) | rev_e);
}



/* Palette initialisation. */

static inline uint16_t rgb8_to_rgb16(uint8_t rgb)
{
	return ((((rgb & 0xe0) >> 5) * 4) << 11) | ((((rgb & 0x1c) >> 2) * 9) << 6) | ((rgb & 0x03) * 10);
}

static inline uint16_t rgb4_to_rgb16(uint8_t rgb)
{
	return ((((rgb & 8) >> 3) * 0x1f) << 11) | ((((rgb & 6) >> 1) * 0x15) << 5) | ((rgb & 1) * 0x1f);
}

static void lcd_init_palette(vidinfo_t *vid)
{
	uint16_t *palette = (uint16_t *) vid->jz_mem.palette;
	uint16_t *end = (uint16_t *) palette + (1 << (vid->jz_fb->bpp));
	uint8_t value = 0;

	while (palette < end)
	{
		switch (vid->jz_fb->bpp)
		{
			case 4:
			*palette = rgb4_to_rgb16(value);
			break;

			case 8:
			default:
			*palette = rgb8_to_rgb16(value);
			break;
		}

		value++;
		palette++;
	}
}



static void jz_lcd_set_timing(vidinfo_t *vid)
{
	uint32_t pclk = lcd_get_pixel_clock(vid);
	uint32_t val;

#ifdef CONFIG_CPU_JZ4730
        val = __cpm_get_pllout() / pclk;
        lcd_ctrl_set(vid, CPM_CFCR2, val - 1);
        val = pclk * 4 ;
        if ( val > 150000000 ) {
                val = 150000000;
        }
        val = __cpm_get_pllout() / val;
        val--;
        if ( val > 0xF )
                val = 0xF;
#else
	int pll_div;

	pll_div = lcd_ctrl_get(vid, CPM_CPCCR) & lcd_ctrl_get(vid, CPM_CPCCR_PCS); /* clock source,0:pllout/2 1: pllout */
	pll_div = pll_div ? 1 : 2 ;
	val = ( __cpm_get_pllout()/pll_div ) / pclk;
	val--;
	if ( val > 0x1ff ) {
		val = 0x1ff;
	}
	__cpm_set_pixdiv(val);

	val = pclk * 3 ;	/* LCDClock > 2.5*Pixclock */
	if ( val > 150000000 ) {
		val = 150000000;
	}
	val = ( __cpm_get_pllout()/pll_div ) / val;
	val--;
	if ( val > 0x1f ) {
		val = 0x1f;
	}
#endif
	__cpm_set_ldiv( val );
	lcd_ctrl_set(vid, CPM_CPCCR, lcd_ctrl_get(vid, CPM_CPCCR) | CPM_CPCCR_CE); /* update divide */
}

/* Initialise the LCD controller with the memory, panel and framebuffer details.
   Return the pixel clock frequency. */

static void jz_lcd_ctrl_init(void *lcd_base, void *fb_vaddr, vidinfo_t *vid)
{
	struct jz_mem_info *fbi = &vid->jz_mem;
	struct jz_fb_dma_descriptor *descriptors = (struct jz_fb_dma_descriptor *) lcd_get_descriptors(vid);

	/* Set the LCD controller address. */

	vid->lcd = lcd_base;

	/* Position framebuffer regions at the given address. */

	fbi->screen = (uint32_t) fb_vaddr;

	/* Obtain the palette address. */

	fbi->palette = lcd_get_palette(vid);

	/* Initialise a palette for lower colour depths. */

	if (vid->jz_fb->bpp < 12)
		lcd_init_palette(vid);

	/* Reference the descriptors in memory. */

	fbi->dmadesc_fb0 = &descriptors[0];
	fbi->dmadesc_fb1 = &descriptors[1];
	fbi->dmadesc_palette = &descriptors[2];

	/* Populate descriptors. */

	/* Provide the first framebuffer descriptor in single and dual modes. */

	fbi->dmadesc_fb0->fsadr = lcd_get_framebuffer(0, vid);
	fbi->dmadesc_fb0->fidr = 0;
	fbi->dmadesc_fb0->ldcmd = lcd_get_size(vid) / 4; /* length in words */

	/* Provide the second framebuffer descriptor only in dual-panel mode. */

	if (lcd_get_panels(vid) == 2)
	{
		fbi->dmadesc_fb1->fdadr = fbi->dmadesc_fb1;
		fbi->dmadesc_fb1->fsadr = lcd_get_framebuffer(1, vid);
		fbi->dmadesc_fb1->fidr = 0;
		fbi->dmadesc_fb1->ldcmd = lcd_get_size(vid) / 4;

		/* Note the address to be provided for the second channel. */

		fbi->fdadr1 = fbi->dmadesc_fb1;
	}

	/* Initialise palette descriptor details if a palette is to be used. */

	/* Assume any mode with <12 bpp is palette driven. */

	if (vid->jz_fb->bpp < 12)
	{
		fbi->dmadesc_palette->fsadr = fbi->palette;
		fbi->dmadesc_palette->fidr = 0;
		fbi->dmadesc_palette->ldcmd = (lcd_get_palette_size(vid) / 4) | LCD_CMD_PAL;

		/* Flip back and forth between the palette and framebuffer. */

		fbi->dmadesc_palette->fdadr = fbi->dmadesc_fb0;
		fbi->dmadesc_fb0->fdadr = fbi->dmadesc_palette;

		/* Provide the palette descriptor address first. */

		fbi->fdadr0 = fbi->dmadesc_palette;
	}
	else
	{
		/* No palette: always use the framebuffer descriptor. */

		fbi->dmadesc_fb0->fdadr = fbi->dmadesc_fb0;
		fbi->fdadr0 = fbi->dmadesc_fb0;
	}
}

/* Initialise the LCD registers. */

static void jz_lcd_hw_init(vidinfo_t *vid)
{
	struct jzfb_info *jzfb = vid->jz_fb;
	uint32_t val = 0;

	/* Compute control register flags. */

	switch (jzfb->bpp) {
		case 1:
		val |= LCD_CTRL_BPP_1;
		break;

		case 2:
		val |= LCD_CTRL_BPP_2;
		break;

		case 4:
		val |= LCD_CTRL_BPP_4;
		break;

		case 8:
		val |= LCD_CTRL_BPP_8;
		break;

		case 15:
		val |= LCD_CTRL_RGB555;
		case 16:
		val |= LCD_CTRL_BPP_16;
		break;

		case 17 ... 32:
		val |= LCD_CTRL_BPP_18_24;	/* target is 4bytes/pixel */
		break;

		default:
		val |= LCD_CTRL_BPP_16;		/* default to 16bpp */
		break;
	}

	/* Set various configuration registers for the panel. */

	switch (jzfb->cfg & MODE_MASK) {
		case MODE_STN_MONO_DUAL:
		case MODE_STN_COLOR_DUAL:
		val |= jz_lcd_stn_init(jzfb->h >> 1, vid);
		break;

		case MODE_STN_MONO_SINGLE:
		case MODE_STN_COLOR_SINGLE:
		val |= jz_lcd_stn_init(jzfb->h, vid);
		break;

		case MODE_TFT_GEN:
		case MODE_TFT_CASIO:
		case MODE_8BIT_SERIAL_TFT:
		case MODE_TFT_18BIT:
		jz_lcd_tft_init(vid);
		break;

		case MODE_TFT_SAMSUNG:
		jz_lcd_samsung_init(vid);
		break;

		case MODE_TFT_SHARP:
		jz_lcd_sharp_init(vid);
		break;

		default:
		break;
	}

	/* Further control register and panel configuration. */

	val |= LCD_CTRL_BST_16;		/* Burst Length is 16WORD=64Byte */
	val |= LCD_CTRL_OFUP;		/* OutFIFO underrun protect */

	lcd_ctrl_set(vid, LCD_CTRL, val);
	lcd_ctrl_set(vid, LCD_CFG, jzfb->cfg);
}

static void jz_lcd_timing_init(vidinfo_t *vid)
{
	__cpm_stop_lcd();
	jz_lcd_set_timing(vid);
	__cpm_start_lcd();
	udelay(1000);
}

/* Initialise DMA for the driver. */

static void jz_lcd_dma_init(vidinfo_t *vid)
{
	struct jz_mem_info *fbi = &vid->jz_mem;
	uint32_t mode = vid->jz_fb->cfg & MODE_MASK;

	/* Configure DMA by setting frame descriptor addresses. */

	lcd_ctrl_set(vid, LCD_DA0, (uint32_t) fbi->fdadr0);

	if ((mode == MODE_STN_COLOR_DUAL) || (mode == MODE_STN_MONO_DUAL))
		lcd_ctrl_set(vid, LCD_DA1, (uint32_t) fbi->fdadr1);
}

/* Set the colour depth. */

void lcd_set_bpp(uint8_t bpp)
{
	vidinfo_t *vid = &panel_info;
	vid->jz_fb->bpp = bpp;
}

void lcd_enable()
{
	vidinfo_t *vid = &panel_info;

	/* Clear the disable bit (DIS) and set the enable bit (ENA). */

	lcd_ctrl_set(vid, LCD_CTRL, (lcd_ctrl_get(vid, LCD_CTRL) & ~LCD_CTRL_DIS) | LCD_CTRL_ENA);
}

void lcd_disable()
{
	vidinfo_t *vid = &panel_info;

	/* Set the disable bit (DIS). */

	lcd_ctrl_set(vid, LCD_CTRL, lcd_ctrl_get(vid, LCD_CTRL) | LCD_CTRL_DIS);
}

void lcd_quick_disable()
{
	vidinfo_t *vid = &panel_info;

	/* Clear the enable bit (ENA) for quick disable. */

	lcd_ctrl_set(vid, LCD_CTRL, lcd_ctrl_get(vid, LCD_CTRL) & ~LCD_CTRL_ENA);
}

uint32_t lcd_ctrl_init()
{
	vidinfo_t *vid = &panel_info;
	struct jz_mem_info *fbi = &vid->jz_mem;
	void *fb_vaddr;

	/* Start from the top of memory and obtain the framebuffer region. */

	fb_vaddr = (void *) lcd_get_framebuffer(0, vid);

	jz_lcd_ctrl_init(0, fb_vaddr, vid);
	flush_cache_all();
	jz_lcd_hw_init(vid);
	jz_lcd_timing_init(vid);
	jz_lcd_dma_init(vid);

	return fbi->screen;
}
